#include <windows.h>
#include <stdio.h>
#include "md5.h"

#define BUFSIZE 0x400

typedef struct HashSearchInfoStruct {
	ULARGE_INTEGER fsMin;
	ULARGE_INTEGER fsMax;
	FILETIME ft;
} HASH_SEARCH_INFO, *PHASH_SEARCH_INFO;

typedef DWORD (WINAPI *GLPN)(LPCTSTR, LPTSTR, DWORD);
typedef BOOL (WINAPI *IWOW64P)(HANDLE, PBOOL);
typedef BOOL (WINAPI *WOW64DWFR)(PVOID*);
typedef BOOL (WINAPI *WOW64RWFR)(PVOID);

static GLPN g_pGetLongPathName = NULL;
static IWOW64P g_pIsWow64Process = NULL;
static WOW64DWFR g_pWow64DisableWow64FsRedirection = NULL;
static WOW64RWFR g_pWow64RevertWow64FsRedirection = NULL;

#define NEWSEARCH 1
#define ONLYFIRST 2
#define NOSUBDIRS 4
#define NOFILECMP 8
#define SIZERANGE 16

#define FindFileSizeInRange(fd, min, max) \
    ( ((min.HighPart < fd.nFileSizeHigh) || (min.HighPart == fd.nFileSizeHigh && min.LowPart <= fd.nFileSizeLow)) && \
      ((max.HighPart > fd.nFileSizeHigh) || (max.HighPart == fd.nFileSizeHigh && max.LowPart >= fd.nFileSizeLow)) \
    )

#define FindFileSizeCompare(fd, fs) \
    ((fs.LowPart == fd.nFileSizeLow && fs.HighPart == fd.nFileSizeHigh))

#define FindFileLastWriteTimeCompare(ft, ftlwt) \
    ((ft.dwLowDateTime == ftlwt.dwLowDateTime && ft.dwHighDateTime == ftlwt.dwHighDateTime))

void MD5ToString(char* string, md5_byte_t checksum[16])
{
	int i = 0;
	int j = 0;
	int a1, a2;
	for(i = 0; i < 16; i++) {
		a1 = (checksum[i] >> 4) % 16;
		a2 = checksum[i] % 16;
		string[j++] = a1 < 10 ? ('0' + a1) : ('a' + a1 - 10);
		string[j++] = a2 < 10 ? ('0' + a2) : ('a' + a2 - 10);
	}
	string[j++] = '\0';
}

BOOL MD5File(HANDLE file, char* hash)
{
	md5_state_t state;
	md5_byte_t digest[16];
	BYTE md5block[BUFSIZE];
	DWORD read;

	md5_init(&state); 

	do {
		if (ReadFile(file, md5block, BUFSIZE, &read, NULL) == FALSE)
			return FALSE;

		md5_append(&state, md5block, read);
	}
	while (read);

	md5_finish(&state, digest);
	MD5ToString(hash, digest);

	return TRUE;
}

BOOL HashFile(const char *filename, char *hash)
{
	HANDLE hFile;

	hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ | 
			FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

	if (hFile == INVALID_HANDLE_VALUE)
	{
		printf("Could not open file: %s\n", filename);
		return FALSE;
	}

	if (!MD5File(hFile, hash))
	{
		printf("Error reading file: %s\n", filename);
		CloseHandle(hFile);
		return FALSE;
	}

	CloseHandle(hFile);

	return TRUE;
}

DWORD pGetLongPathName(LPCTSTR lpszShortPath, LPTSTR lpszLongPath, DWORD cchBuffer)
{
	if (g_pGetLongPathName)
		return g_pGetLongPathName(lpszShortPath, lpszLongPath, cchBuffer);

	return 0;
}

BOOL pIsWow64Process(HANDLE hProcess, PBOOL Wow64Process)
{
	if (g_pIsWow64Process)
		return g_pIsWow64Process(hProcess, Wow64Process);

	return FALSE;
}

BOOL pWow64DisableWow64FsRedirection(PVOID *OldValue)
{
	if (g_pWow64DisableWow64FsRedirection)
		return g_pWow64DisableWow64FsRedirection(OldValue);

	return FALSE;
}

BOOL pWow64RevertWow64FsRedirection(PVOID OldValue)
{
	if (g_pWow64RevertWow64FsRedirection)
		return g_pWow64RevertWow64FsRedirection(OldValue);

	return FALSE;
}

void LoadFunctionEntryPoints()
{
	HMODULE hModule;
	hModule = GetModuleHandle("kernel32.dll");
	
	if (hModule != NULL)
	{
		g_pGetLongPathName = (GLPN)
				GetProcAddress(hModule, "GetLongPathNameA");

		g_pIsWow64Process = (IWOW64P)
				GetProcAddress(hModule, "IsWow64Process");

		g_pWow64DisableWow64FsRedirection = (WOW64DWFR)
				GetProcAddress(hModule, "Wow64DisableWow64FsRedirection");

		g_pWow64RevertWow64FsRedirection = (WOW64RWFR)
				GetProcAddress(hModule, "Wow64RevertWow64FsRedirection");
	}
}

void HashSearch(const char *filename, const char *hash, char *directory, PHASH_SEARCH_INFO phsi, UINT flags)
{
	HANDLE h;
	WIN32_FIND_DATA fd;
	static int matchcount = 0;
	static UINT fileount = 0;
	static BOOL isWow64 = FALSE;
	static PVOID OldValue = NULL;
	char buf[MAX_PATH];
	char *ptr;

	ptr = (char *)directory;
	while (*ptr != '\0') ptr++;
	while (*(ptr-1) == '\\') ptr--;
	lstrcpy(ptr, "\\*.*");

	if (flags & NEWSEARCH)
	{
		printf("Searching for files with a MD5 checksum of %s\n\n", hash);

		// Disable fs redirection on 64-bit Windows
		pIsWow64Process(GetCurrentProcess(), &isWow64);

		if (isWow64)
			pWow64DisableWow64FsRedirection(&OldValue);

		if (filename == NULL)
			flags |= NOFILECMP;

		matchcount = 0;
		fileount = 0;
	}

	h = FindFirstFile(directory, &fd);
	if (h != INVALID_HANDLE_VALUE)
	{
		do
		{
			if (fd.cFileName[0] != '.' || (fd.cFileName[1] != '\0' && 
					(fd.cFileName[1] != '.' || fd.cFileName[2] != '\0')))
			{
				lstrcpy(ptr + 1, fd.cAlternateFileName[0] != '\0' ? fd.cAlternateFileName : fd.cFileName);

				if ((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
				{
					fileount++;
					if (((flags & SIZERANGE) && FindFileSizeInRange(fd, phsi->fsMin, phsi->fsMax)) || 
							((flags & SIZERANGE) == 0 && 
									((flags & NOFILECMP) == NOFILECMP || 
									(FindFileSizeCompare(fd, phsi->fsMin) == TRUE && 
											(FindFileLastWriteTimeCompare(phsi->ft, fd.ftLastWriteTime) == FALSE || 
											lstrcmpi(directory, filename) != 0)
									))
							))
					{
						if (HashFile(directory, buf) == FALSE)
							continue;

						if (lstrcmpi(buf, hash) == 0)
						{
							matchcount++;

							if (flags & ONLYFIRST)
								fileount = 0xFFFFFFFF;

							if (pGetLongPathName(directory, buf, MAX_PATH) == 0)
								lstrcpy(buf, directory);

							printf("%s\n", buf);
						}
					}
				}
				else
				{
					if ((flags & NOSUBDIRS) == 0)
						HashSearch(filename, hash, directory, phsi, (flags & ~NEWSEARCH));
				}
			}
		}
		while (FindNextFile(h,&fd) && fileount != 0xFFFFFFFF);

		FindClose(h);
	}
	else
	{
		if (flags & NEWSEARCH || (GetLastError() != ERROR_ACCESS_DENIED))
		{
			ptr[0] = '\0'; // fix directory
			if (pGetLongPathName(directory, buf, MAX_PATH) == 0)
				lstrcpy(buf, directory);

			printf("Error searching directory: %s\\*.*\n", buf);
		}

		if (flags & NEWSEARCH)
		{
			if (isWow64)
				pWow64RevertWow64FsRedirection(OldValue);

			return;
		}
	}

	ptr[0] = '\0'; // fix directory

	if (flags & NEWSEARCH)
	{
		printf("\nSearch completed, found %d file(s) with a matching checksum.\n", matchcount);

		if (isWow64)
			pWow64RevertWow64FsRedirection(OldValue);
	}
}

BOOL IsValidHash(const char *hash)
{
	LPCTSTR p = hash;

	if (strlen(hash) != 32)
		return FALSE;

	while (*p) 
	{
		if ((*p < '0' || *p > '9') && 
			(*p < 'a' || *p > 'f') && 
			(*p < 'A' || *p > 'F'))
			return FALSE;
		p++;
	}
	return TRUE;
}

void sizetollu(const char *ptr, PULARGE_INTEGER fs)
{
	while (*ptr == ' ') ptr++;

	if (*ptr == '-')
	{
		fs->LowPart = 0xFFFFFFFF;
		fs->HighPart = 0xFFFFFFFF;
		return;
	}

	fs->LowPart = 0;
	fs->HighPart = 0;

	while (*ptr != '\0')
	{
		if (*ptr < '0' || *ptr > '9') break;
		fs->QuadPart *= 10;
		fs->QuadPart += (*ptr - '0');
		ptr++;
	}

	if (ptr[0] != '\0' && ptr[1] == 'B' && 
			(ptr[2] == ',' || ptr[2] == '\0' || ptr[2] == ' '))
	{
		if (ptr[0] == 'K') fs->QuadPart *= 1024;
		if (ptr[0] == 'M') fs->QuadPart *= 1048576;
		if (ptr[0] == 'G') fs->QuadPart *= 1073741824;
	}
}

void GetSizeRange(char *ptr, PULARGE_INTEGER min, PULARGE_INTEGER max)
{
	sizetollu(ptr, min);

	if (min->LowPart == 0xFFFFFFFF && 
			min->HighPart == 0xFFFFFFFF)
	{
		min->LowPart = 0;
		min->HighPart = 0;
	}

	while (*ptr != '\0' && *ptr != ',')
		ptr++;

	if (*ptr == ',')
	{
		ptr++;
		sizetollu(ptr, max);
	}
	else
	{
		max->LowPart = min->LowPart;
		max->HighPart = min->HighPart;
	}
}

void help()
{
	printf("Usage: md5search.exe <md5_hash|filename> [-f][-s][-n][-r filesize] <directory>\n\n");
	printf("   -f     Stop searching after first matching file.\n");
	printf("   -s     Do not recurse into subdirectories.\n");
	printf("   -n     Do not compare file sizes when using filename.\n");
	printf("   -r [size|min,max]\n");
	printf("          Only search for files that are within a size range.\n");
	printf("          The range value supports (KB, MB, GB) size units and\n");
	printf("          when using min,max range you can use - for min or\n");
	printf("          max value to search for any files that are at most\n");
	printf("          max size or at least min size.\n");
}

int main(int argc, char **argv)
{
	printf("\nMD5 search 0.5 - Search files by MD5 checksum\n");
	printf("Copyright (C) 2011 - 2013 Pasi Ruokola\n");
	printf("Website - http://koti.mbnet.fi/vaultec/\n\n");
	
	LoadFunctionEntryPoints();

	if (argc > 2)
	{
		char directory[1024];
		char filename[MAX_PATH];
		char *filenameptr = NULL;
		BOOL isWow64 = FALSE;
		PVOID OldValue = NULL;
		char hash[33];
		UINT searchflags;
		HANDLE hFile;
		HASH_SEARCH_INFO hsi;
		int i;

		searchflags = NEWSEARCH;
		hsi.fsMax.LowPart = INVALID_FILE_SIZE;
		hsi.fsMax.HighPart = INVALID_FILE_SIZE;

		if (IsValidHash(argv[1]) == TRUE)
		{
			searchflags |= NOFILECMP;
			lstrcpy(hash, argv[1]);

			hsi.fsMin.LowPart = INVALID_FILE_SIZE;
			hsi.fsMin.HighPart = INVALID_FILE_SIZE;

			hsi.ft.dwLowDateTime = 0xFFFFFFFF;
			hsi.ft.dwHighDateTime = 0xFFFFFFFF;
		}
		else
		{
			// Disable fs redirection on 64-bit Windows
			pIsWow64Process(GetCurrentProcess(), &isWow64);

			if (isWow64)
				pWow64DisableWow64FsRedirection(&OldValue);

			hFile = CreateFile(argv[1], GENERIC_READ, FILE_SHARE_READ | 
					FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);

			if (hFile == INVALID_HANDLE_VALUE)
			{
				printf("Could not open file: %s\n", argv[1]);
				return EXIT_FAILURE;
			}

			hsi.fsMin.LowPart = GetFileSize(hFile, &hsi.fsMin.HighPart);
			GetFileTime(hFile, NULL, NULL, &hsi.ft);

			if (GetShortPathName(argv[1], filename, MAX_PATH) == 0)
				lstrcpy(filename, argv[1]);

			if (isWow64)
				pWow64RevertWow64FsRedirection(OldValue);

			filenameptr = filename;

			if (!MD5File(hFile, hash))
			{
				printf("Error reading file: %s\n", filename);
				CloseHandle(hFile);
				return EXIT_FAILURE;
			}

			CloseHandle(hFile);
		}

		for (i = 2; i < argc; i++)
		{
			if ((argv[i][0] == '-' || 
					argv[i][0] == '/') && 
					argv[i][2] == '\0')
			{
				if (argv[i][1] == '?')
				{
					help();
					break;
				}
				else if (argv[i][1] == 'f') searchflags |= ONLYFIRST;
				else if (argv[i][1] == 's') searchflags |= NOSUBDIRS;
				else if (argv[i][1] == 'n') searchflags |= NOFILECMP;
				else if (argv[i][1] == 'r')
				{
					if ((i + 1) >= argc)
						return EXIT_FAILURE;

					searchflags |= SIZERANGE;
					searchflags |= NOFILECMP;

					GetSizeRange(argv[i + 1], &hsi.fsMin, &hsi.fsMax);
					i++;
				}
				else
				{
					printf("Invalid parameter: %s\n\n", argv[i]);
					help();
					return EXIT_FAILURE;
				}
			}
			else
			{
				if (GetShortPathName(argv[i], directory, MAX_PATH) == 0)
					lstrcpy(directory, argv[i]);

				HashSearch(filenameptr, hash, directory, &hsi, searchflags);
			}
		}
	}
	else
	{
		help();
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
